part of dart_mason;

class Texture extends NativeFieldWrapperClass2 {
  final int width;
  final int height;
  final TextureFormat format;

  /// Loads a [Texture] from a url. Only jpegs are supported.
  static Future<Texture> FromUrl(String url) async {
    return new Future<Texture>(() async {
      print("Texture From URL ${url}");
      var client = new io.HttpClient();
      io.HttpResponse response = await client.get(url);


      var bytes = response.body.readBytes();
      List<List<int>> allBytesList = [];
      var totalSize = 0;
      var completer = new Completer<Texture>();
      bytes.listen((a){
        totalSize += a.length;
        allBytesList.add(a);
      }).onDone((){
        _InternalBuffer buffer = new _InternalBuffer();
        buffer.initialize(totalSize);
        var start = 0;
        for(var l in allBytesList){
          var end = start + l.length;
          buffer.loadList(start, l);
          start = end;
        }

        if(response.headers.contentType.mimeType == "image/jpeg"){
          var start = new DateTime.now();
          var decoder = new JPEGDecoder();
          decoder.convertAsync(buffer.toList()).then((texture){
            completer.complete(texture);
            var end = new DateTime.now();
            print("JPG: ${end.difference(start)}");
          });
        }
        if(response.headers.contentType.mimeType == "image/png"){
          var start = new DateTime.now();
          var image = decodePng(buffer.toList());
          var texture = new Texture(image.width, image.height);
          texture.SetPixelData(new Uint8List.view(image.data.buffer));
          completer.complete(texture);
          var end = new DateTime.now();
          print("PNG: ${end.difference(start)}");
        }
      });
      return completer.future;
    });
  }

  Texture(this.width, this.height, {this.format = TextureFormat.RGBA}) {
    _constructor(this.width, this.height, _textureFormatMap[this.format], true);
  }

  Texture._(this.width, this.height, this.format) {
    _constructor(this.width, this.height, _textureFormatMap[this.format], true);
  }

  SetPixelData(Uint8List data) {
    SetPixelDataNative(0, 0, width, height, data);
  }

  SetPixelDataNative(int x, int y, int width, int height, Uint8List data)
      native "Texture_SetPixelData";

  _constructor(int width, int height, int format, bool mipmap)
      native "Texture_constructor";
}

enum TextureFormat {
  RGB,
  RGBA,
  RED,
  RG,
}

final _textureFormatMap = {
  TextureFormat.RGB: 0x1907,
  TextureFormat.RED: 0x1903,
  TextureFormat.RG: 0x8227,
  TextureFormat.RGBA: 0x1908,
};
