part of dart_mason;

typedef void OnBeginFrame(FrameData data);

class UniformList {
  UniformList();
}

class UniformValue {
  final String key;
  final dynamic value;
  final UniformType type;
  final int _typeId;

  UniformValue(this.key, this.value, this.type) : _typeId = type.index {}

  String toString() {
    return "[${key}: - ${value}]";
  }
}

class RenderableNode {
  List<Renderable> renderables = [];
  List<RenderableNode> children = [];
}

class Renderable extends NativeFieldWrapperClass2 {
  final materials.MaterialInstance _material;
  final Mesh _mesh;
  final Matrix4 _transform;

  Renderable(this._material, this._mesh, this._transform) {
    _material.params;
    _material.material.params;
    _material.material.textures;

    var params = new Map<String, Object>();

    _material.material.params.forEach((String value, d){
      params[d.uniformKey] = d.defaultValue;
    });

    _material.material.textures.forEach((String key, d){
      params[d.uniformKey] = d.defaultValue;
    });

    _material.params.forEach((String key, Object value){
      if(_material.material.params[key] != null){
        params[_material.material.params[key].uniformKey] = value;
      }
    });

    _constructor(_material, _mesh, _transform, params.length);

    params.forEach((String key, Object value){
      _setUniform(key, _getType(value), value);
    });

  }

  _constructor(materials.MaterialInstance material, Mesh mesh, Matrix4 transform, int uniform_size)
      native "Renderable_constructor";

  _setUniform(String key, int type, Object value) native "Renderable_SetUniform";
  _setLight(Matrix4 transform) native "Renderable_SetLight";

  int _getType(Object value){
    if(value is Vector4){
      return UniformType.VEC4.index;
    }
    if(value is Vector3){
      return UniformType.VEC3.index;
    }
    if(value is Vector2){
      return UniformType.VEC2.index;
    }
    if(value is double){
      return UniformType.FLOAT.index;
    }
    if(value is Texture){
      return UniformType.TEXTURE.index;
    }
    return -1;
  }

}

class World extends NativeFieldWrapperClass2 {
  Scene _current;

  int _frameIndex = 0;

  OnBeginFrame onBeginFrame;

  World._() {
    _constructor();
    Image image = new Image(10,10);
  }

  _constructor() native "World_constructor";

  setScene(Scene scene) {
    _current = scene;
    _setScene(scene);
  }

  render(List<Renderable> renderList, [List<_Light> lights = const []]){
    var max_lights = 8;
    var lights_array = [];
    if(lights.length > max_lights){
      throw new Exception("No more than 8 lights allowed.");
    }
    lights.forEach((_Light light){
      if(light is PointLight){
        lights_array.add([light.position.clone(), light.color.clone()]);
      }
    });
    //fill the rest of the lights with "black" lights.
    var black_lights = max_lights - lights.length;
    for(var i = 0; i < black_lights; i++){
      //position origin and no color (black)
      lights_array.add([new Vector3(0.0, 0.0, 0.0), new Vector3(0.0, 0.0, 0.0)]);
    }
   _render_internal(lights_array, renderList);
  }

  _render_internal(List<List> lights, List<Renderable> renderList) native "World_Render";

  /// Prepares a
  renderScene(Scene scene) {
    var renderList = new List<Renderable>();
    _processComponents(scene._components, renderList, new Matrix4.identity());
    render(renderList);
  }

  _processComponents(List<Component> components, List<Renderable> renderList,
      Matrix4 parentTransform) {
    for (var c in components) {
      var newTransform = c._computeTransform(parentTransform);
      for (var mesh in c._meshes) {
        renderList.add(new Renderable(1, mesh,
            [new UniformValue("model", newTransform, UniformType.MAT4)]));
      }
      _processComponents(c._components, renderList, newTransform);
    }
  }

  _setScene(Scene scene) native "World_setScene";

  void scheduleFrame() native "World_ScheduleFrame";
}

class Headset extends TrackableDevice {
  Headset._(Vector3 position, Quaternion orientation)
      : super._(position, orientation);
}

class TrackableDevice extends Device {
  final Vector3 position;
  final Quaternion orientation;

  TrackableDevice._(this.position, this.orientation);
}

class Device {
  static const OCULUS_TOUCH_RIGHT = "oculus_touch_right";
  static const OCULUS_TOUCH_LEFT = "oculus_touch_left";
}

class FrameData {
  final Duration timeStamp;
  final Duration timeDelta;
  final Headset headset;
  final Map<String, TrackableDevice> devices;

  FrameData._(this.timeStamp, this.timeDelta, this.headset, this.devices) {}
}

Duration _previousTime;

_beginFrame(int microseconds, List<double> headsetData,
    Map<String, List<double>> deviceData, int frameIndex) {
  var missedFrames = frameIndex - world._frameIndex - 1;
  world._frameIndex = frameIndex;
  if (missedFrames > 0) {
    print("Missed Frames: ${missedFrames}");
  }
  if (world.onBeginFrame != null) {
    var duration = new Duration(microseconds: microseconds);
    Duration deltaTime;
    if(_previousTime == null){
      deltaTime = new Duration();
    } else {
      deltaTime = duration - _previousTime;
    }
    _previousTime = duration;
    var headset = new Headset._(
        new Vector3.array(headsetData), new Quaternion.array(headsetData, 3));
    var devices = new Map<String, TrackableDevice>();
    deviceData.forEach((key, values) {
      var device = new TrackableDevice._(
          new Vector3.array(values), new Quaternion.array(values, 3));
      devices[key] = device;
    });
    var frameData = new FrameData._(duration, deltaTime, headset, devices);
    world.onBeginFrame(frameData);
  }
}

final World world = new World._();
