// Copyright (c) 2013, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

part of solarium.io;

// Global constants.
class _Const {
    // Bytes for "HTTP".
    static const HTTP = const [72, 84, 84, 80];
    // Bytes for "HTTP/1.".
    static const HTTP1DOT = const [72, 84, 84, 80, 47, 49, 46];
    // Bytes for "HTTP/1.0".
    static const HTTP10 = const [72, 84, 84, 80, 47, 49, 46, 48];
    // Bytes for "HTTP/1.1".
    static const HTTP11 = const [72, 84, 84, 80, 47, 49, 46, 49];

    static const bool T = true;
    static const bool F = false;
    // Loopup-map for the following characters: '()<>@,;:\\"/[]?={} \t'.
    static const SEPARATOR_MAP = const [
        F,F,F,F,F,F,F,F,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,T,F,T,F,F,
        F,F,F,T,T,F,F,T,F,F,T,F,F,F,F,F,F,F,F,F,F,T,T,T,T,T,T,T,F,F,F,F,F,F,F,F,F,
        F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,T,T,T,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
        F,F,F,F,F,F,F,F,F,F,F,F,T,F,T,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
        F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
        F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,
        F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F,F];
}


// Frequently used character codes.
class _CharCode {
    static const int HT = 9;
    static const int LF = 10;
    static const int CR = 13;
    static const int SP = 32;
    static const int AMPERSAND = 38;
    static const int COMMA = 44;
    static const int DASH = 45;
    static const int SLASH = 47;
    static const int ZERO = 48;
    static const int ONE = 49;
    static const int COLON = 58;
    static const int SEMI_COLON = 59;
    static const int EQUAL = 61;
}


// States of the HTTP parser state machine.
class _State {
    static const int START = 0;
    static const int METHOD_OR_RESPONSE_HTTP_VERSION = 1;
    static const int RESPONSE_HTTP_VERSION = 2;
    static const int REQUEST_LINE_METHOD = 3;
    static const int REQUEST_LINE_URI = 4;
    static const int REQUEST_LINE_HTTP_VERSION = 5;
    static const int REQUEST_LINE_ENDING = 6;
    static const int RESPONSE_LINE_STATUS_CODE = 7;
    static const int RESPONSE_LINE_REASON_PHRASE = 8;
    static const int RESPONSE_LINE_ENDING = 9;
    static const int HEADER_START = 10;
    static const int HEADER_FIELD = 11;
    static const int HEADER_VALUE_START = 12;
    static const int HEADER_VALUE = 13;
    static const int HEADER_VALUE_FOLDING_OR_ENDING = 14;
    static const int HEADER_VALUE_FOLD_OR_END = 15;
    static const int HEADER_ENDING = 16;

    static const int CHUNK_SIZE_STARTING_CR = 17;
    static const int CHUNK_SIZE_STARTING_LF = 18;
    static const int CHUNK_SIZE = 19;
    static const int CHUNK_SIZE_EXTENSION = 20;
    static const int CHUNK_SIZE_ENDING = 21;
    static const int CHUNKED_BODY_DONE_CR = 22;
    static const int CHUNKED_BODY_DONE_LF = 23;
    static const int BODY = 24;
    static const int CLOSED = 25;
    static const int UPGRADED = 26;
    static const int FAILURE = 27;

    static const int FIRST_BODY_STATE = CHUNK_SIZE_STARTING_CR;
}

// HTTP version of the request or response being parsed.
class _HttpVersion {
    static const int UNDETERMINED = 0;
    static const int HTTP10 = 1;
    static const int HTTP11 = 2;
}

// States of the HTTP parser state machine.
class _MessageType {
    static const int UNDETERMINED = 0;
    static const int REQUEST = 1;
    static const int RESPONSE = 0;
}


/**
 * The _HttpDetachedStreamSubscription takes a subscription and some extra data,
 * and makes it possible to "inject" the data in from of other data events
 * from the subscription.
 *
 * It does so by overriding pause/resume, so that once the
 * _HttpDetachedStreamSubscription is resumed, it'll deliver the data before
 * resuming the underlaying subscription.
 */
class _HttpDetachedStreamSubscription implements StreamSubscription<List<int>> {
    StreamSubscription<List<int>> _subscription;
    List<int> _injectData;
    bool _isCanceled = false;
    int _pauseCount = 1;
    Function _userOnData;
    bool _scheduled = false;

    _HttpDetachedStreamSubscription(this._subscription,
        this._injectData,
        this._userOnData);

    bool get isPaused => _subscription.isPaused;

    Future<T> asFuture<T>([T futureValue]) =>
        _subscription.asFuture<T>(futureValue);

    Future cancel() {
        _isCanceled = true;
        _injectData = null;
        return _subscription.cancel();
    }

    void onData(void handleData(List<int> data)) {
        _userOnData = handleData;
        _subscription.onData(handleData);
    }

    void onDone(void handleDone()) {
        _subscription.onDone(handleDone);
    }

    void onError(Function handleError) {
        _subscription.onError(handleError);
    }

    void pause([Future resumeSignal]) {
        if (_injectData == null) {
            _subscription.pause(resumeSignal);
        } else {
            _pauseCount++;
            if (resumeSignal != null) {
                resumeSignal.whenComplete(resume);
            }
        }
    }

    void resume() {
        if (_injectData == null) {
            _subscription.resume();
        } else {
            _pauseCount--;
            _maybeScheduleData();
        }
    }

    void _maybeScheduleData() {
        if (_scheduled) return;
        if (_pauseCount != 0) return;
        _scheduled = true;
        scheduleMicrotask(() {
            _scheduled = false;
            if (_pauseCount > 0 || _isCanceled) return;
            var data = _injectData;
            _injectData = null;
            // To ensure that 'subscription.isPaused' is false, we resume the
            // subscription here. This is fine as potential events are delayed.
            _subscription.resume();
            if (_userOnData != null) {
                _userOnData(data);
            }
        });
    }
}


class _HttpDetachedIncoming extends Stream<List<int>> {
    final StreamSubscription<List<int>> subscription;
    final List<int> bufferedData;

    _HttpDetachedIncoming(this.subscription, this.bufferedData);

    StreamSubscription<List<int>> listen(void onData(List<int> event),
        {Function onError,
            void onDone(),
            bool cancelOnError}) {
        if (subscription != null) {
            subscription
                ..onData(onData)
                ..onError(onError)
                ..onDone(onDone);
            if (bufferedData == null) {
                return subscription..resume();
            }
            return new _HttpDetachedStreamSubscription(subscription,
                bufferedData,
                onData)
                ..resume();
        } else {
            // TODO(26379): add test for this branch.
            return new Stream<List<int>>.fromIterable([bufferedData])
                .listen(onData,
                onError: onError,
                onDone: onDone,
                cancelOnError: cancelOnError);
        }
    }
}


/**
 * HTTP parser which parses the data stream given to [consume].
 *
 * If an HTTP parser error occours, the parser will signal an error to either
 * the current _HttpIncoming or the _parser itself.
 *
 * The connection upgrades (e.g. switching from HTTP/1.1 to the
 * WebSocket protocol) is handled in a special way. If connection
 * upgrade is specified in the headers, then on the callback to
 * [:responseStart:] the [:upgrade:] property on the [:HttpParser:]
 * object will be [:true:] indicating that from now on the protocol is
 * not HTTP anymore and no more callbacks will happen, that is
 * [:dataReceived:] and [:dataEnd:] are not called in this case as
 * there is no more HTTP data. After the upgrade the method
 * [:readUnparsedData:] can be used to read any remaining bytes in the
 * HTTP parser which are part of the protocol the connection is
 * upgrading to. These bytes cannot be processed by the HTTP parser
 * and should be handled according to whatever protocol is being
 * upgraded to.
 */
class _HttpParser {

    static bool _isTokenChar(int byte) {
        return byte > 31 && byte < 128 && !_Const.SEPARATOR_MAP[byte];
    }

    static bool _isValueChar(int byte) {
        return (byte > 31 && byte < 128) || (byte == _CharCode.SP) ||
            (byte == _CharCode.HT);
    }

    static List<String> _tokenizeFieldValue(String headerValue) {
        List<String> tokens = new List<String>();
        int start = 0;
        int index = 0;
        while (index < headerValue.length) {
            if (headerValue[index] == ",") {
                tokens.add(headerValue.substring(start, index));
                start = index + 1;
            } else if (headerValue[index] == " " || headerValue[index] == "\t") {
                start++;
            }
            index++;
        }
        tokens.add(headerValue.substring(start, index));
        return tokens;
    }

    static int _toLowerCaseByte(int x) {
        // Optimized version:
        //  -  0x41 is 'A'
        //  -  0x7f is ASCII mask
        //  -  26 is the number of alpha characters.
        //  -  0x20 is the delta between lower and upper chars.
        return (((x - 0x41) & 0x7f) < 26) ? (x | 0x20) : x;
    }

}
